import logging
import os
import sys

import click


GREY = "\x1b[38;20m"
YELLOW = "\x1b[33;20m"
RED = "\x1b[31;20m"
BOLD_RED = "\x1b[31;1m"
RESET = "\x1b[0m"


class CustomFormatter(logging.Formatter):
    def format(self, record):
        log_fmt = "%(asctime)s - %(levelname)s - %(message)s"
        formats = {
            logging.DEBUG: f"{GREY}{log_fmt}{RESET}",
            logging.INFO: f"{GREY}{log_fmt}{RESET}",
            logging.WARNING: f"{YELLOW}{log_fmt}{RESET}",
            logging.ERROR: f"{RED}{log_fmt}{RESET}",
            logging.CRITICAL: f"{BOLD_RED}{log_fmt}{RESET}",
        }
        log_fmt = formats.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def get_handler():
    try:
        log_file = "/var/log/urnc.log"
        h = logging.FileHandler(log_file)
        print(f"Logging to {log_file}")
        return h
    except Exception:
        pass
    try:
        log_file = os.path.expanduser("~/.urnc.log")
        h = logging.FileHandler(log_file)
        print(f"Logging to {log_file}")
        return h
    except Exception:
        pass
    return None


def setup_logger(use_file=True, verbose=False):
    logger = logging.getLogger(__name__)
    if verbose:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)
    if (use_file):
        handler = get_handler()
        if (handler is not None):
            logger.addHandler(handler)
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setFormatter(CustomFormatter())

    logger.addHandler(stdout_handler)


def dbg(msg):
    logger = logging.getLogger(__name__)
    logger.debug(msg)


def log(msg):
    logger = logging.getLogger(__name__)
    logger.info(msg)


def warn(msg):
    logger = logging.getLogger(__name__)
    logger.warning(msg)


def error(msg):
    logger = logging.getLogger(__name__)
    logger.error(msg)


def critical(msg):
    logger = logging.getLogger(__name__)
    logger.critical(msg)
    raise Exception(msg)



#!/usr/bin/env python3
import click
import os

from urnc.convert import check, convert
from urnc.edit import edit
from urnc.student import student
from urnc.version import version
from urnc.pull import pull
from urnc.ci import ci


@click.group(help="urnc manages UR FIDS courses")
@click.option("-f", "--root", help="The course folder", default=os.getcwd())
@click.pass_context
def main(ctx, root):
    ctx.ensure_object(dict)
    ctx.obj["ROOT"] = root
    pass


main.add_command(version)
main.add_command(convert)
main.add_command(ci)
main.add_command(check)
main.add_command(edit)
main.add_command(student)
main.add_command(pull)



import click

import semver

import urnc.util as util


def bump(version, action):
    v = semver.Version.parse(version)
    print(v)
    match action:
        case "show":
            return None
        case "patch":
            return v.bump_patch()
        case "minor":
            return v.bump_minor()
        case "major":
            return v.bump_major()
    raise click.UsageError("Invalid action")



def version_self(ctx, action, repo):
    config = util.read_pyproject(ctx)
    v = config["project"]["version"]
    new_version = bump(v, action)
    if not new_version:
        return 
    if repo is not None and repo.is_dirty():
        print(f"Repo is not clean, commit your changes before calling version")
        return 
    print(f"    New Version: {new_version}")
    config["project"]["version"] = str(new_version)
    path = util.write_pyproject(ctx, config)
    if repo is not None:
        message = f"v{new_version}"
        print("Committing new version and tagging the commit...")
        repo.index.add(path)
        repo.index.commit(message)
        repo.create_tag(message, "HEAD", message)
        print("Done.")



def version_course(ctx, action, repo):
    config = util.read_config(ctx)
    v = config["version"]
    new_version = bump(v, action)
    if not new_version:
        return 
    if repo is not None and repo.is_dirty():
        print(f"Repo is not clean, commit your changes before calling version")
        return
    print(f"    New Version: {new_version}")
    config["version"] = str(new_version)
    path = util.write_config(ctx, config)
    if repo is not None:
        message = f"v{new_version}"
        print("Committing new version and tagging the commit...")
        repo.index.add(path)
        repo.index.commit(message)
        repo.create_tag(message, "HEAD", message)
        print("Done.")


def get_repo(ctx, action, no_git):
    if(no_git or action == "show"):
        return None
    return util.get_git_repo(ctx)


@click.command(help="Manage the semantic version of your course")
@click.option("--no-git", is_flag=True, help="Do not make a git commit")
@click.option("--self", is_flag=True, help="Echo the version of urnc")
@click.argument(
    "action",
    type=click.Choice(["show", "patch", "minor", "major"]),
    required=False,
    default="show",
)
@click.pass_context
def version(ctx, no_git, self, action):
    repo = get_repo(ctx, action, no_git)
    if self:
        version_self(ctx, action, repo)
    else:
        version_course(ctx, action, repo)



import inspect
import re
from nbconvert.preprocessors.base import Preprocessor

import urnc.preprocessor.util as util


def replace_local_link(cell, verbose):
    if cell.cell_type != "markdown":
        return
    if not re.search(r"\(#Exercise-.*\)", cell.source, re.IGNORECASE):
        return
    if verbose:
        print("Detected Link to exercise.")
    cell.source = re.sub(r"#Exercise-", "#ex-", cell.source, re.IGNORECASE)


class ProcessExercises(Preprocessor):
    def preprocess(self, notebook, resources):
        verbose = resources["verbose"]
        for cell in notebook.cells:
            replace_local_link(cell, verbose)
            if util.has_tag(cell, util.Tags.EXERCISE_START):
                eid = cell.metadata.exercise_id
                html = f"""
                    <a class="anchor" name="ex-{eid}"></a>
                    <h3>   ✎    Exercise {eid}</h3>
                """
                cell.source = inspect.cleandoc(html)
        return notebook, resources



import re
import click


import urnc.util as util
import urnc.student as student

@click.command(help="Edit the course, switch to main branch")
@click.pass_context
def edit(ctx):
    repo = util.get_git_repo(ctx)

    main_branch = "main"
    if not util.branch_exists(repo, main_branch):
        repo.git.remote("set-branches","--add", "origin", "main")

    repo.git.fetch("--all")
    repo.git.switch(main_branch)
    idx = student.find_urnc_stash(repo)
    if(idx):
        repo.git.stash("pop", idx)




import click
import re


import urnc.util as util
from urnc.convert import convert_fn


def find_urnc_stash(repo):
    stash = repo.git.stash("list").split("\n")
    pattern = r'stash@{(?P<idx>\d+)}:[^:]+:\s+urnc'
    for line in stash:
        match = re.match(pattern, line)
        if match:
            return match.group("idx")
    return None




@click.command(help="Build and show the student version")
@click.pass_context
def student(ctx):
    repo = util.get_git_repo(ctx)
    if repo.active_branch.name != "main":
        raise Exception("You are not in the main branch")
    existing = find_urnc_stash(repo)
    if(existing):
        raise Exception("There are already stashed changes for urnc")

    repo.git.stash("push", "-m", "urnc")
    idx = find_urnc_stash(repo)
    if(not idx):
        raise Exception("Failed to stash changes properly")
    repo.git.checkout("-B", "urnc_tmp")
    repo.git.stash("apply", idx)
    convert_fn(ctx,
               input=repo.working_dir,
               output=repo.working_dir,
               force=True,
               verbose=True,
               dry_run=False
    )
    repo.git.add(all=True)
    repo.index.commit("urnc test convert")




import click
import os
import git


import urnc.util as util

import urnc.logger as log

base_url_admin = "https://git.uni-regensburg.de/fids/"
base_url = "https://git.spang-lab.de/"


def get_upstream_changes(repo):
    branch_name = repo.active_branch
    changes = repo.git.diff(
        f"..origin/{branch_name}", "--name-status").split('\n')
    return changes


def get_upstream_deleted(repo):
    changes = get_upstream_changes(repo)
    files = []
    for change in changes:
        [ctype, file] = change.split('\t', 1)
        if (ctype == "D"):
            files.append(file)
    return files


def get_upstream_added(repo):
    changes = get_upstream_changes(repo)
    files = []
    for change in changes:
        [ctype, file] = change.split('\t', 1)
        if (ctype == "A"):
            files.append(file)
    return files


def get_remote_url(isAdmin, course_name, user, token):
    if not isAdmin:
        return f"https://git.spang-lab.de/{course_name}"
    return f"https://{user}:{token}@git.uni-regensburg.de/fids/{course_name}"


def get_folder_name(isAdmin, course_name, output):
    folder_name = output
    if output is None:
        folder_name = course_name
    if isAdmin:
        folder_name = f"{folder_name}-admin"
    return folder_name


def get_repo(folder_name):
    try:
        git_repo = git.Repo(folder_name)
        return git_repo
    except Exception:
        return log.critical(f"{folder_name} is not a git repo")


def reset_deleted_files(repo):
    branch = repo.active_branch
    deleted_files = repo.git.ls_files("--deleted").split('\n')
    deleted_upstream = get_upstream_deleted(repo)
    for filename in deleted_files:
        if (filename in deleted_upstream):
            repo.git.checkout("HEAD", '--', filename)
        else:
            log.dbg(f"Restoring deleted file {filename}")
            repo.git.checkout(f"origin/{branch}", "--", filename)


def merge(repo):
    branch = repo.active_branch


@ click.command(help="Pull the course repo")
@ click.argument(
    "course_name",
    type=str,
    required=True


)
@ click.option("-o", "--output", type=str, help="The name of the output folder", default=None)
@ click.option("-b", "--branch", help="The branch to pull", default="main")
@ click.option("-d", "--depth", help="The depth for git fetch", default=1)
@ click.option("-t", "--token", help="The git access token", required=True)
@ click.option("-u", "--user", help="The name of the access token", default="urnc")
@ click.pass_context
def pull(ctx, course_name, output, branch, depth, token, user):
    log.setup_logger()
    isAdmin = os.getenv('JUPYTERHUB_ADMIN_ACCESS') == "1"

    folder_name = get_folder_name(isAdmin, course_name, output)
    git_url = get_remote_url(isAdmin, course_name, user, token)

    if (not os.path.exists(folder_name)):
        log.log(f"{folder_name} does not exists. Cloning repo {git_url}")
        try:
            git.Repo.clone_from(git_url, folder_name,
                                branch=branch, depth=depth)
            log.log("Cloned successfully.")
        except Exception as err:
            log.error("Failed to clone repo. Error:")
            log.error(err)
        return
    log.log(f"{folder_name} exists. Updating the repo")
    repo = get_repo(folder_name)

    log.log(f"Fetching changes...")
    repo.remote().fetch()
    if (isAdmin):
        log.log(f"Trying to pull")
        repo.git.pull("--ff-only")
        log.log(f"Pulled repo. Done.")
        return

    log.log(f"Restoring locally deleted files")
    reset_deleted_files(repo)
    log.log(f"unstaging all changes")
    repo.git.reset("--mixed")

    if repo.is_dirty():
        log.log("Repo is dirty. Commiting....")
        repo.git.commit("-am", "Automatic commit by urnc",
                        "-c", "user.email=urnc@spang-lab.de",
                        "-c", "user.name=urnc",
                        "--allow-empty")
        log.log("Created new commit")

    log.log("Merging from remote...")
    merge(repo)
    log.log("Done.")



from nbconvert.preprocessors.base import Preprocessor

import urnc.preprocessor.util as util
from urnc.preprocessor.util import Tags, Keywords
import re


def process_comments(cell):
    text = cell.source
    lines = text.split("\n")
    last_tag = None
    found_tag = False
    tlines = []
    for line in lines:
        if re.match(Keywords.SOLUTION, line, re.IGNORECASE):
            last_tag = Tags.SOLUTION
            found_tag = True
            continue
        if re.match(Keywords.SKELETON, line, re.IGNORECASE):
            last_tag = Tags.SKELETON
            found_tag = True
            continue
        if re.match(Keywords.SOLUTION_END, line, re.IGNORECASE):
            last_tag = None
            continue
        tlines.append((line, last_tag))
    if not found_tag:
        return None
    slines = []
    for line, tag in tlines:
        if tag is None:
            slines.append(line)
        if tag == Tags.SOLUTION:
            continue
        if tag == Tags.SKELETON:
            uncom = re.sub(r"^#\s?", "", line)
            slines.append(uncom)
    if len(slines) == 0:
        return None
    cell.source = "\n".join(slines)
    return cell


class RemoveSolutions(Preprocessor):
    def preprocess(self, notebook, resources):
        cells = []
        for cell in notebook.cells:
            if util.has_tag(cell, Tags.SOLUTION):
                cell = process_comments(cell)
            if cell is not None:
                cells.append(cell)
        notebook.cells = cells
        return notebook, resources



from nbconvert.preprocessors.base import Preprocessor

import re
import urnc.preprocessor.util as util
from urnc.preprocessor.util import Keywords, Tags


def critical_error(text):
    RED = "\033[1;31m"
    RESET = "\033[0;0m"
    print("%sCritical Error: %s %s" % (RED, text, RESET))


def end_exercise(exercise_id, has_solution):
    if exercise_id is None:
        return
    if not has_solution:
        critical_error("Exercise '%s' has no solution" % exercise_id)


class AddTags(Preprocessor):
    def preprocess(self, notebook, resources):
        verbose = resources["verbose"]

        exercise_ids = set()
        exercise_id = None
        has_solution = False

        for cell in notebook.cells:
            if m := re.match(Keywords.EXERCISE_DEPRECATED, cell.source, re.IGNORECASE):
                print("'Exercise' as Keyword is deprecated. Use 'Assignment' instead")
            if match := re.search(Keywords.EXERCISE_START, cell.source, re.IGNORECASE):
                end_exercise(exercise_id, has_solution)
                has_solution = False
                exercise_id = match.group(1)
                util.set_tag(cell, Tags.EXERCISE_START)
                if exercise_id in exercise_ids:
                    critical_error("Duplicate Exercise id '%s'" % exercise_id)
                else:
                    exercise_ids.add(exercise_id)
                if verbose:
                    print("Detected Exercise '%s'" % exercise_id)
            elif util.has_header(cell) and exercise_id is not None:
                end_exercise(exercise_id, has_solution)
                exercise_id = None
            if exercise_id is None:
                continue
            util.set_tag(cell, Tags.EXERCISE)
            cell.metadata.exercise_id = exercise_id

            if re.search(Keywords.SOLUTION, cell.source, re.IGNORECASE) or util.has_tag(
                cell, Tags.SOLUTION
            ):
                if verbose:
                    print(" Detected Solution cell %s" %
                          util.cell_preview(cell))
                util.set_tag(cell, Tags.SOLUTION)
                has_solution = True
        end_exercise(exercise_id, has_solution)
        return notebook, resources
